/**
 * 
 */
package com.midimind.core.hardware;

import java.util.ArrayList;
import java.util.List;

import com.midimind.core.SystemProperties;
import com.midimind.core.hardware.command.Command;
import com.midimind.core.hardware.command.PayloadCommand;

/**
 * @author Davide
 *
 */
public abstract class Device {

	private static final byte MIDI_SYSEX_START = (byte)0xF0;
	private static final byte MIDI_SYSEX_END = (byte)0xF7;
	private static final byte DEVICE_ACK = (byte)0x0E;

	public static Device[] getAvailableDevices() throws DeviceException {
		String id = SystemProperties.getProperties().getProperty("device");
		String pack = Device.class.getPackage().getName() + '.' + id.toLowerCase();
		
		Device device = null;
		
		try {
			device = (Device)Class.forName(pack + "." + id + "Device").newInstance();
		} catch (Exception e) {
			throw new DeviceException("No device implementation found for \'" + id + "\'",
					DeviceState.UNKNOWN_DEVICE_IMPLEMENTATION);
		}
		
		return device.getAvailableDevices_internal();
	}
	
	public abstract String getDeviceID();
	
	public abstract void open() throws DeviceException;
	
	public abstract void close() throws DeviceException;
	
	public abstract void reset() throws DeviceException;
	
	public abstract boolean isOpen() throws DeviceException;
	
	protected abstract int read() throws DeviceException;
	
	protected abstract void write(byte b) throws DeviceException;
	
	protected abstract Device[] getAvailableDevices_internal()
			throws DeviceException;	
	
	public byte[] send(Command command) throws DeviceException {
		byte sysExID = new Byte(SystemProperties.getProperties().getProperty("sysex_id"));
		write(MIDI_SYSEX_START);
		write(sysExID);
		write(command.getHeader());
		if (command instanceof PayloadCommand)
			for (byte b : ((PayloadCommand)command).getPayload())
				write(b);
		write(MIDI_SYSEX_END);
		
		List<Byte> buffer = new ArrayList<Byte>();
			
		if (read() != MIDI_SYSEX_START)
			throw new DeviceException("Miss midi system exclusive message start.",
					DeviceState.MALFORMED_RESPONSE);
		if (read() != sysExID)
			throw new DeviceException("System exclusive id not match.",
					DeviceState.MALFORMED_RESPONSE);
			
		int b = 0;	
		do {
			b = read();
			if (b == -1)
				throw new DeviceException("Miss midi system exclusive message end.",
						DeviceState.MALFORMED_RESPONSE);
			buffer.add((byte)b);
		} while (b != MIDI_SYSEX_END);
			
		byte[] bytebuffer = new byte[buffer.size() - 1];
		for (int i = 0; i < buffer.size() - 1; i++)
			bytebuffer[i] = buffer.get(i);
			
		if (command.hasResponse()) {
			return bytebuffer;
		} else {
			if ((bytebuffer.length == 0) || (bytebuffer[0] != DEVICE_ACK))
				throw new DeviceException("Miss device ack.",
						DeviceState.MALFORMED_RESPONSE);
			return null;
		}		
	}
	
	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		else if (!(obj instanceof Device))
			return false;
		else {
			Device other = (Device)obj;
			return other.getDeviceID().equals(this.getDeviceID());
		}
	}

	@Override
	public int hashCode() {
		return this.getDeviceID().hashCode();
	}
	
}
